/******************************************************************************
 *
 * Program: CPU Load NT plugin for Nagios (nrpe_nt).
 * License: GPL
 *
 * Designed as a Nagios Plugin (based on plugins by Ethan Galstad).
 * Copyright (C) 2004 Stephen Strudwick (Pipex Communications, sas@pipex.net)
 *
 * License Information:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

/* A class that extracts the current CPU load and places it into a history queue.
   Operations can then be done on the data in that queue to find the mean, variance
   and standard deviation of the data 

   CPU load is extracted by a query of NTDLL.DLL via the function NtQuerySystemInformation.

   Win NT/2000/XP only.
*/

#include "stdafx.h"
#include "cpuusage.h"

/* Initialiase the class, grab the current idle and system time and create the queue */
cpuusage::cpuusage()
{
  SYSTEM_PERFORMANCE_INFORMATION SysPerfInfo;
  SYSTEM_TIME_INFORMATION        SysTimeInfo;
  LONG                           status;

  NtQuerySystemInformation = (PROCNTQSI)GetProcAddress(
                                        GetModuleHandle("ntdll"),
                                        "NtQuerySystemInformation");


  Q = CreateQueue(MAX_QUEUE_HISTORY);

  if (!NtQuerySystemInformation)
    return;

  // get new system time
  status = NtQuerySystemInformation(SystemTimeInformation,&SysTimeInfo,sizeof(SysTimeInfo),0);
  if (status!=NO_ERROR)
    return;

  // get new CPU's idle time
  status = NtQuerySystemInformation(SystemPerformanceInformation,&SysPerfInfo,sizeof(SysPerfInfo),NULL);
  if (status != NO_ERROR)
    return;

  // store CPU's idle and system time
  liOldIdleTime = SysPerfInfo.liIdleTime;
  liOldSystemTime = SysTimeInfo.liKeSystemTime;

}

/* Destroying it would be nice */
cpuusage::~cpuusage()
{
  DisposeQueue(Q);
}

/* find out the average CPU load since the idle/system time was last set and add it to the queue */
void cpuusage::poll() {
  SYSTEM_PERFORMANCE_INFORMATION SysPerfInfo;
  SYSTEM_TIME_INFORMATION        SysTimeInfo;
  SYSTEM_BASIC_INFORMATION       SysBaseInfo;
  double                         dbIdleTime;
  double                         dbSystemTime;
  LONG                           status;

  NtQuerySystemInformation = (PROCNTQSI)GetProcAddress(
                                        GetModuleHandle("ntdll"),
                                        "NtQuerySystemInformation");

  if (!NtQuerySystemInformation)
    return;

  // get number of processors in the system
  status = NtQuerySystemInformation(SystemBasicInformation,&SysBaseInfo,sizeof(SysBaseInfo),NULL);
  if (status != NO_ERROR)
    return;
    
  // get new system time
  status = NtQuerySystemInformation(SystemTimeInformation,&SysTimeInfo,sizeof(SysTimeInfo),0);
  if (status!=NO_ERROR)
    return;

  // get new CPU's idle time
  status = NtQuerySystemInformation(SystemPerformanceInformation,&SysPerfInfo,sizeof(SysPerfInfo),NULL);
  if (status != NO_ERROR)
    return;

  if (liOldIdleTime.QuadPart != 0)
  {
     // CurrentValue = NewValue - OldValue
     dbIdleTime = Li2Double(SysPerfInfo.liIdleTime) - Li2Double(liOldIdleTime);
     dbSystemTime = Li2Double(SysTimeInfo.liKeSystemTime) - Li2Double(liOldSystemTime);

     // CurrentCpuIdle = IdleTime / SystemTime
     dbIdleTime = dbIdleTime / dbSystemTime;

     //dbIdleTime = 100.0 - (dbIdleTime * 100.0) / (double)SysBaseInfo.bKeNumberProcessors + 0.5;
	 dbIdleTime = 100.0 - (dbIdleTime * 100.0) / (double)SysBaseInfo.bKeNumberProcessors;

	if ((UINT)dbIdleTime >= 0 || (UINT)dbIdleTime <= 100) {
	 // Also add to memory
	 if (IsFull(Q))
	   Dequeue(Q);
	 Enqueue((UINT)dbIdleTime, Q);
	}
   }

  // store new CPU's idle and system time
  liOldIdleTime = SysPerfInfo.liIdleTime;
  liOldSystemTime = SysTimeInfo.liKeSystemTime;
}

/* Find the mean of the data in the queue 
   max_history is how far back in the queues history to look (i.e last 60 entries)

   Note: these two functions are very hacky, but I dont have
   time to make them nice */

float cpuusage::mean(int max_history) {
  int n, top, stop;
  int count = 0;
  float total = 0;

  if (max_history < 0) /* No minus numbers */
    return 0;

  top = Rear(Q);
  if (top > 0) { /* Make sure we have a valid size */
	if (top > max_history) /* Work out how far back to look */
	  stop = top-max_history;
	else
	  stop = 1;

    /* Must traverse queue backwards, as newest data is at back of queue */
    for(n=top;n>=stop;n--) {
	  count++; /* Count how many elements we have */
	        total += Element(n,Q);
	}
	if (count > 0)  /* Catch Divide by Zero */
	  return total / count;
	else
	  return 0;
  } else
	return 0;
}

/* Find the variance of the data in the queue
   max_history is how far back in the queues history to look (i.e last 60 entries) */
float cpuusage::variance(int max_history, float cpu_mean) {
  int n, top, stop;
  int count = 0;
  float total = 0;

  if (max_history < 0)
    return 0;

  top = Rear(Q);

  if (top > 0) {

	if (top > max_history)
	  stop = top-max_history;
	else
	  stop = 1;

    /* Must traverse queue backwards, as newest data is at back of queue */
    for(n=top;n>=stop;n--) {
	  count++;
      total += (Element(n,Q) - cpu_mean) * (Element(n,Q) - cpu_mean);
	}
	if (count > 0)  /* Catch Divide by Zero */
	  return total / count;
	else
	  return 0;
  } else
	return 0;
}
